<?php

namespace Drupal\eca_entity_print\Plugin\Action;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Component\Transliteration\TransliterationInterface;
use Drupal\Core\Access\AccessManagerInterface;
use Drupal\Core\Action\ConfigurableActionBase;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Token\TokenInterface;
use Drupal\entity_print\Plugin\ExportTypeManagerInterface;
use Drupal\entity_print\PrintBuilderInterface;
use Drupal\file\Entity\File;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Print file document from entity render.
 *
 * @Action(
 *   id = "eca_print_file_from_entity",
 *   label = @Translation("Print file document from entity"),
 *   description = @Translation("Print file document from a chosen entity and saves it as a file entity."),
 *   type = "entity"
 * )
 */
class PrintFileFromEntity extends ConfigurableActionBase implements ContainerFactoryPluginInterface {

  /**
   * Access manager.
   *
   * @var \Drupal\Core\Access\AccessManagerInterface
   */
  protected $accessManager;

  /**
   * The Token services.
   *
   * @var \Drupal\eca\Token\TokenInterface
   */
  protected TokenInterface $tokenServices;

  /**
   * The Print builder service.
   *
   * @var \Drupal\entity_print\PrintBuilderInterface
   */
  protected $printBuilder;

  /**
   * The Entity Print plugin manager.
   *
   * @var \Drupal\entity_print\Plugin\EntityPrintPluginManagerInterface
   */
  protected $entityPrintPluginManager;

  /**
   * The export type manager.
   *
   * @var \Drupal\entity_print\Plugin\ExportTypeManagerInterface
   */
  protected $exportTypeManager;

  /**
   * The transliteration service.
   *
   * @var \Drupal\Component\Transliteration\PhpTransliteration
   */
  protected $transliteration;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, AccessManagerInterface $access_manager, TokenInterface $token_services, PrintBuilderInterface $print_builder, PluginManagerInterface $entity_print_plugin_manager, ExportTypeManagerInterface $export_type_manager, TransliterationInterface $transliteration) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->accessManager = $access_manager;
    $this->tokenServices = $token_services;
    $this->printBuilder = $print_builder;
    $this->entityPrintPluginManager = $entity_print_plugin_manager;
    $this->exportTypeManager = $export_type_manager;
    $this->transliteration = $transliteration;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('access_manager'),
      $container->get('eca.token_services'),
      $container->get('entity_print.print_builder'),
      $container->get('plugin.manager.entity_print.print_engine'),
      $container->get('plugin.manager.entity_print.export_type'),
      $container->get('transliteration')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE) {
    /** @var \Drupal\Core\Entity\EntityInterface $object */
    $route_params = [
      'export_type' => $this->configuration['export_type'],
      'entity_id' => $object->id(),
      'entity_type' => $object->getEntityTypeId(),
    ];
    return $this->accessManager->checkNamedRoute('entity_print.view', $route_params, $account, $return_as_object);
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function execute($entity = NULL) {
    if (!($entity instanceof FieldableEntityInterface)) {
      return;
    }

    $print_engine = $this->entityPrintPluginManager->createSelectedInstance($this->getExportType());
    $file_name = $this->getFileName() . '.' . $print_engine->getExportType()->getFileExtension();

    $uri = $this->printBuilder->savePrintable([$entity], $print_engine, 'private', $file_name);

    if ($uri) {
      /** @var \Drupal\file\Entity\File $file */
      $file = File::create([
        'filename' => $file_name,
        'uri' => $uri,
        'status' => 1,
        'uid' => 1,
      ]);
      $file->setPermanent();
      $file->save();

      $token_name = trim($this->configuration['token_name']);
      if ($token_name === '') {
        $token_name = 'eca-entity-output-filename';
      }
      $this->tokenServices->addTokenData($token_name, $file);
    }

  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'token_name' => '',
      'export_type' => 'pdf',
      'file_name' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $form['token_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Name of token'),
      '#default_value' => $this->configuration['token_name'],
      '#description' => $this->t('Name of the token available after file document from view render is generated. Contains the file entity.'),
      '#weight' => -60,
    ];

    $form['file_name'] = [
      '#type' => 'textarea',
      '#title' => $this->t('The exported file name'),
      '#description' => $this->t('Provide a name for the generated file. Do not use "/" in the name. This property supports tokens.'),
      '#default_value' => $this->configuration['file_name'],
      '#required' => TRUE,
      '#weight' => -20,
    ];

    $form['export_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Export type'),
      '#options' => $this->exportTypeManager->getFormOptions(),
      '#required' => TRUE,
      '#default_value' => $this->configuration['export_type'],
      '#weight' => -10,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['token_name'] = $form_state->getValue('token_name');
    $this->configuration['file_name'] = $form_state->getValue('file_name');
    $this->configuration['export_type'] = $form_state->getValue('export_type');
  }

  /**
   * Get the configured document export type.
   *
   * @return string
   *   The document export type.
   */
  protected function getExportType(): string {
    return trim((string) $this->tokenServices->replaceClear($this->configuration['export_type']));
  }

  /**
   * Get the configured file name.
   *
   * @return string
   *   The file name.
   */
  protected function getFileName(): string {
    if (!empty($this->configuration['file_name'])) {
      $file_name = trim((string) $this->tokenServices->replaceClear($this->configuration['file_name']));
      return $this->sanitizeFilename($file_name);
    }
    return uniqid('eca.view.output', TRUE);
  }

  /**
   * Sanitize the file name.
   *
   * @param string $file_name
   *   The file name that will be sanitized.
   *
   * @return string
   *   Sanitized file name.
   */
  protected function sanitizeFileName(string $file_name) {
    $file_name = $this->transliteration->transliterate($file_name);
    // Replace whitespace.
    $file_name = str_replace(' ', '-', $file_name);
    // Remove remaining unsafe characters.
    $file_name = preg_replace('![^0-9A-Za-z_.-]!', '', $file_name);
    // Remove multiple consecutive non-alphabetical characters.
    $file_name = preg_replace('/(_)_+|(\.)\.+|(-)-+/', '\\1\\2\\3', $file_name);
    // Force lowercase to prevent issues on case-insensitive file systems.
    $file_name = strtolower($file_name);

    return $file_name;
  }

}
