<?php

namespace Drupal\eca_entity_print\Plugin\Action;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Component\Transliteration\TransliterationInterface;
use Drupal\Core\Access\AccessManagerInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Action\ConfigurableActionBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Token\TokenInterface;
use Drupal\entity_print\Plugin\ExportTypeManagerInterface;
use Drupal\entity_print\PrintBuilderInterface;
use Drupal\file\Entity\File;
use Drupal\views\Entity\View;
use Drupal\views\ViewExecutable;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Print file document from view render.
 *
 * @Action(
 *   id = "eca_print_file_from_views",
 *   label = @Translation("Print file document from views output"),
 *   description = @Translation("Print a file document from a chosen views output and saves it as a file entity."),
 * )
 */
class PrintFileFromViews extends ConfigurableActionBase implements ContainerFactoryPluginInterface {

  /**
   * Access manager.
   *
   * @var \Drupal\Core\Access\AccessManagerInterface
   */
  protected $accessManager;

  /**
   * The Token services.
   *
   * @var \Drupal\eca\Token\TokenInterface
   */
  protected TokenInterface $tokenServices;

  /**
   * The Print builder service.
   *
   * @var \Drupal\entity_print\PrintBuilderInterface
   */
  protected $printBuilder;

  /**
   * The Entity Print plugin manager.
   *
   * @var \Drupal\entity_print\Plugin\EntityPrintPluginManagerInterface
   */
  protected $entityPrintPluginManager;

  /**
   * The export type manager.
   *
   * @var \Drupal\entity_print\Plugin\ExportTypeManagerInterface
   */
  protected $exportTypeManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The transliteration service.
   *
   * @var \Drupal\Component\Transliteration\PhpTransliteration
   */
  protected $transliteration;

  /**
   * The executable view.
   *
   * @var \Drupal\views\ViewExecutable
   */
  protected ViewExecutable $view;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, AccessManagerInterface $access_manager, TokenInterface $token_services, PrintBuilderInterface $print_builder, PluginManagerInterface $entity_print_plugin_manager, ExportTypeManagerInterface $export_type_manager, EntityTypeManagerInterface $entity_type_manager, TransliterationInterface $transliteration) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->accessManager = $access_manager;
    $this->tokenServices = $token_services;
    $this->printBuilder = $print_builder;
    $this->entityPrintPluginManager = $entity_print_plugin_manager;
    $this->exportTypeManager = $export_type_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->transliteration = $transliteration;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('access_manager'),
      $container->get('eca.token_services'),
      $container->get('entity_print.print_builder'),
      $container->get('plugin.manager.entity_print.print_engine'),
      $container->get('plugin.manager.entity_print.export_type'),
      $container->get('entity_type.manager'),
      $container->get('transliteration')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE) {
    $result = AccessResult::forbidden();
    if (($display = $this->view->getDisplay()) && $display->access($account)) {
      $result = AccessResult::allowed();
    }
    return $return_as_object ? $result : $result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($object = NULL): void {

    if (!$this->getView() || !isset($this->view)) {
      return;
    }
    /** @var \Drupal\views\Entity\View $view */
    $view = $this->getView();

    $print_engine = $this->entityPrintPluginManager->createSelectedInstance($this->getExportType());
    $file_name = $this->getFileName() . '.' . $print_engine->getExportType()->getFileExtension();

    $uri = $this->printBuilder->savePrintable([$view], $print_engine, 'private', $file_name);

    if ($uri) {
      /** @var \Drupal\file\Entity\File $file */
      $file = File::create([
        'filename' => $file_name,
        'uri' => $uri,
        'status' => 1,
        'uid' => 1,
      ]);
      $file->setPermanent();
      $file->save();

      $token_name = trim($this->configuration['token_name']);
      if ($token_name === '') {
        $token_name = 'eca-view-output-filename';
      }
      $this->tokenServices->addTokenData($token_name, $file);
    }

  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'token_name' => '',
      'view_id' => '',
      'display_id' => 'default',
      'arguments' => '',
      'export_type' => 'pdf',
      'file_name' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $views = [];

    /** @var \Drupal\views\Entity\View[] $all_views */
    $all_views = $this->entityTypeManager->getStorage('view')->loadMultiple();
    foreach ($all_views as $view) {
      if ($view->status()) {
        $views[$view->id()] = $view->label();
      }
    }
    $form['token_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Name of token'),
      '#default_value' => $this->configuration['token_name'],
      '#description' => $this->t('Name of the token available after file document from view render is generated. Contains the file entity.'),
      '#weight' => -60,
    ];
    $form['view_id'] = [
      '#type' => 'select',
      '#title' => $this->t('View'),
      '#default_value' => $this->configuration['view_id'],
      '#description' => $this->t('Select the view from the list. The view  will always return a list of complete entities.'),
      '#weight' => -50,
      '#options' => $views,
      '#required' => TRUE,
    ];
    $form['display_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Display'),
      '#default_value' => $this->configuration['display_id'],
      '#description' => $this->t('Write the view <code>display id</code> to execute. Set as default to use the default view configuration.'),
      '#weight' => -40,
    ];
    $form['arguments'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Arguments'),
      '#default_value' => $this->configuration['arguments'],
      '#description' => $this->t('Provide the Contextual filters of the view in order and one by line. This property supports tokens.'),
      '#weight' => -30,
    ];

    $form['file_name'] = [
      '#type' => 'textarea',
      '#title' => $this->t('The exported file name'),
      '#description' => $this->t('Provide a name for the generated file. Do not use "/" in the name. This property supports tokens.'),
      '#default_value' => $this->configuration['file_name'],
      '#required' => TRUE,
      '#weight' => -20,
    ];

    $form['export_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Export type'),
      '#options' => $this->exportTypeManager->getFormOptions(),
      '#required' => TRUE,
      '#default_value' => $this->configuration['export_type'],
      '#weight' => -10,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['token_name'] = $form_state->getValue('token_name');
    $this->configuration['view_id'] = $form_state->getValue('view_id');
    $this->configuration['display_id'] = $form_state->getValue('display_id');
    $this->configuration['arguments'] = $form_state->getValue('arguments');
    $this->configuration['file_name'] = $form_state->getValue('file_name');
    $this->configuration['export_type'] = $form_state->getValue('export_type');
  }

  /**
   * Prepare and return view with optional arguments applied.
   *
   * The view gets loaded, its executable set to $this->view and the given
   * display being prepared with optional arguments.
   */
  protected function getView(): EntityInterface|View|null {
    $view_id = $this->getViewId();
    if ($view_id === '') {
      return NULL;
    }

    /** @var \Drupal\views\Entity\View $view */
    $view = $this->entityTypeManager->getStorage('view')->load($view_id);
    if (!$view || !$view->status()) {
      return NULL;
    }

    $this->view = $view->getExecutable();

    $display_id = $this->getDisplayId();
    if ($display_id !== '') {
      if (!$this->view->setDisplay($display_id)) {
        return NULL;
      }
    }
    else {
      $this->view->initDisplay();
    }

    $args = [];
    foreach (explode('/', $this->getArguments()) as $argument) {
      if ($argument !== '') {
        $args[] = $argument;
      }
    }
    $this->view->setArguments($args);

    return $view;
  }

  /**
   * Get the configured token data.
   *
   * @return string
   *   The markup.
   */
  protected function getTokenData(): string {
    return trim((string) $this->tokenServices->replaceClear($this->configuration['token_data']));
  }

  /**
   * Get the configured view ID.
   *
   * @return string
   *   The view ID.
   */
  protected function getViewId(): string {
    return trim((string) $this->tokenServices->replaceClear($this->configuration['view_id']));
  }

  /**
   * Get the configured display ID.
   *
   * @return string
   *   The display ID.
   */
  protected function getDisplayId(): string {
    return trim((string) $this->tokenServices->replaceClear($this->configuration['display_id']));
  }

  /**
   * Get the configured Views arguments.
   *
   * @return string
   *   The arguments, multiple arguments are seraparated by "/".
   */
  protected function getArguments(): string {
    return trim((string) $this->tokenServices->replaceClear($this->configuration['arguments']));
  }

  /**
   * Get the configured document export type.
   *
   * @return string
   *   The document export type.
   */
  protected function getExportType(): string {
    return trim((string) $this->tokenServices->replaceClear($this->configuration['export_type']));
  }

  /**
   * Get the configured file name.
   *
   * @return string
   *   The file name.
   */
  protected function getFileName(): string {
    if (!empty($this->configuration['file_name'])) {
      $file_name = trim((string) $this->tokenServices->replaceClear($this->configuration['file_name']));
      return $this->sanitizeFilename($file_name);
    }
    return uniqid('eca.view.output', TRUE);
  }

  /**
   * Sanitize the file name.
   *
   * @param string $file_name
   *   The file name that will be sanitized.
   *
   * @return string
   *   Sanitized file name.
   */
  protected function sanitizeFileName(string $file_name) {
    $file_name = $this->transliteration->transliterate($file_name);
    // Replace whitespace.
    $file_name = str_replace(' ', '-', $file_name);
    // Remove remaining unsafe characters.
    $file_name = preg_replace('![^0-9A-Za-z_.-]!', '', $file_name);
    // Remove multiple consecutive non-alphabetical characters.
    $file_name = preg_replace('/(_)_+|(\.)\.+|(-)-+/', '\\1\\2\\3', $file_name);
    // Force lowercase to prevent issues on case-insensitive file systems.
    $file_name = strtolower($file_name);

    return $file_name;
  }

}
